# A program which manages Ubuntu-it's web test server
# Copyright (C) 2015-2016  Pietro Albini <pietroalbini@ubuntu.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; witout even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import subprocess
import os
import shutil
import json


BRANCH_DIR_VERSION = "4"


class Branch:
    """Representation of a branch"""

    def __init__(self, manager, name, merge_request):
        self.manager = manager

        self.name = name
        self.mr = None
        self.mr_id = merge_request
        self.mr_url = None
        self.author = None
        self.author_url = None
        self.assignee = None
        self.assignee_url = None
        self.config = None

        self._deploying = False

        self.pinned = name in manager.config["keep-branches"]

        self.check_remote_status()
        self.load_config()

    def check_remote_status(self):
        """Check the branch merging status on GitLab"""
        if self.pinned:
            self.active = True
            return

        result = self.manager.gitlab.merge_request(self.mr_id)
        if result:
            active = result["state"] == "opened"
        else:
            active = False
        self.active = active
        self.mr = result["id"]
        self.mr_url = (self.manager.config["gitlab-url"]+"/" +
                       self.manager.config["gitlab-project"] +
                       "/merge_requests/"+str(self.mr_id))

        self.author = result["author"]["username"]
        self.author_url = result["author"]["web_url"]

        if result["assignee"] is not None:
            self.assignee = result["assignee"]["username"]
            self.assignee_url = result["assignee"]["web_url"]

    # Metadata:

    def present(self):
        """Check if a branch is present"""
        root = self.manager.root / "branches" / self.name
        if root.exists():
            # The branch must be a valid directory
            if not (root.exists() and root.is_dir()):
                return False

            # All the files must be be present
            files = ["version", "commit", "built", "managetests.json"]
            for file in files:
                file = root / file
                if not (file.exists() and file.is_file()):
                    return False

            # The directory version must be correct
            with (root / "version").open() as f:
                if f.read().strip() != BRANCH_DIR_VERSION:
                    return False

            return True

        return False

    def is_running(self):
        """Check if a branch is running"""
        return self.manager.instances.is_running(self.name)

    def has_build_log(self):
        """Check if the branch has a build log"""
        path = self.manager.root / "branches" / self.name / "build.log"
        return path.exists()

    def status(self):
        """Get the status of the branch"""
        if self._deploying:
            return "deploying"
        elif not self.present():
            return "not_present"
        elif self.is_running():
            return "running"
        else:
            return "broken"

    # Deploy and destroy:

    def load_config(self):
        """Load the configuration of this branch"""
        file = self.manager.root / "branches" / self.name / "managetests.json"

        self.config = None

        # Load the configuration only if this branch exists
        if file.exists():
            with file.open() as f:
                self.config = json.load(f)

    def commit(self):
        """Load the commit of this branch"""
        file = self.manager.root / "branches" / self.name / "commit"

        if file.exists():
            if hasattr(self, "_gitlab_commit_cache"):
                delattr(self, "_gitlab_commit_cache")

            with file.open() as f:
                return f.read().strip()
        else:
            # Ask GitLab about the commit
            if not hasattr(self, "_gitlab_commit_cache"):
                self._gitlab_commit_cache = self.manager.gitlab \
                                                .branch_commit(self.name)
            return self._gitlab_commit_cache

    def commit_url(self):
        """Get the URL of a commit"""
        commit = self.commit()
        if not commit:
            return None

        return "%s/%s/commit/%s" % (
            self.manager.config["gitlab-url"],
            self.manager.config["gitlab-project"],
            commit,
        )

    def exec_command(self, raw_command, replaces=None, build=False):
        """Exec a command in the context of the branch"""
        args, kwargs = self.command_popen_args(raw_command, replaces, build)
        return subprocess.check_call(*args, **kwargs)

    def command_popen_args(self, raw_command, replaces=None, build=False):
        """Get the Popen arguments for exec_command"""
        base = self.manager.root / "branches" / self.name

        home = base
        if build:
            home /= "build"

        env = dict(os.environ)
        env["HOME"] = str(home)

        if replaces is None:
            replaces = {}
        replaces.setdefault("base", str(base))

        command = []
        for part in raw_command:
            for key, value in replaces.items():
                part = part.replace("{{%s}}" % key, value)
            command.append(part)

        kwargs = {"env": env, "cwd": str(home)}

        if build:
            log_path = base / "build.log"

            # Log standard output and error
            kwargs["stdin"] = subprocess.DEVNULL
            kwargs["stdout"] = log_path.open("a")
            kwargs["stderr"] = subprocess.STDOUT

        return [command], kwargs

    def deploy(self):
        """Deploy the branch"""
        # Ok, branch already deployed
        if self.present():
            return

        self._deploying = True

        print("[i] Started a deploy of the '%s' branch" % self.name)

        # Start from a fresh branch dir
        branch = self.manager.root / "branches" / self.name
        if branch.exists():
            shutil.rmtree(str(branch))
        branch.mkdir()

        git_url = self.manager.config["git-url"]
        log_file = branch / "build.log"

        # Clear the log file
        log_file.open("w").close()

        def log(msg):
            """Utility function to log a message to the build log"""
            with log_file.open("a") as f:
                f.write(msg+"\n")

        # Create the version file
        with (branch / "version").open("w") as f:
            f.write("%s\n" % BRANCH_DIR_VERSION)

        # Update the bare git repository and checkout the branch
        subprocess.call([
            "git", "clone", "--depth", "1", "--branch", self.name, git_url,
            str(branch / "build"),
        ], stdout=log_file.open("a"), stderr=subprocess.STDOUT)

        # Get the commit of the branch
        commit_file = branch / "commit"
        subprocess.call([
            "git", "--git-dir", str(branch / "build" / ".git"), "rev-parse",
            self.name
        ], stdout=commit_file.open("w"), stderr=log_file.open("a"))

        self.update_status("running")

        # Copy and load the configuration file
        if not (branch / "build" / "managetests.json").exists():
            log("[!] No managetests.json found in the branch!")
            log("[!] Aborting!")
            return self.fail_deployment(branch)

        shutil.copy(
            str(branch / "build" / "managetests.json"),
            str(branch / "managetests.json")
        )
        self.load_config()

        # Build the website
        for cmd in self.config["build"]:
            try:
                self.exec_command(cmd, build=True)
            except (IOError, subprocess.CalledProcessError) as e:
                log("[!] Failed to execute build command: %s" % " ".join(cmd))
                log("[!] Error: %s" % e)
                log("[!] Aborting!")
                return self.fail_deployment(branch)

        # Check if the branch was successifully built
        popen_args = self.command_popen_args(self.config["run"])
        if os.path.exists(popen_args[0][0][0]):
            log("[!] Info: build seems to be successful")
            self.update_status("success")
        else:
            log("[!] Warning: build seems to have failed")
            return self.fail_deployment(branch)

        self.manager.details["branches"][self.name] = self.mr
        self.manager.save_details()

        # Create the built file
        (branch / "built").touch()

        self._deploying = False

    def destroy(self):
        """Destroy the local copy"""
        # Ok, branch not present
        if not self.present():
            return

        shutil.rmtree(
            str(self.manager.root / "branches" / self.name),
            ignore_errors=True
        )

        # Delete the commit cache
        if hasattr(self, "_gitlab_commit_cache"):
            delattr(self, "_gitlab_commit_cache")

        try:
            del self.manager.details["branches"][self.name]
        except KeyError:
            pass
        self.manager.save_details()

    def fail_deployment(self, branch):
        self._deploying = False
        self.update_status('failed')
        self.manager.details["branches"][self.name] = self.mr
        self.manager.save_details()
        # Create the built file
        (branch / "built").touch()
        return False

    def update_status(self, state, log=True):
        """Update the commit status"""
        # Don't update status hooks on development instances
        if not self.manager.config["update-commit-status"]:
            return

        if log:
            url = "%s/+logs/%s/build.log" % (
                self.manager.config["base-url"], self.name
            )
        else:
            url = None

        # Update the commit status in GitLab
        self.manager.gitlab.commit_status(
            self.commit(), state, self.name, url
        )
